package net.knarcraft.serverlauncher.profile;

import net.knarcraft.serverlauncher.server.AdvancedServerType;
import net.knarcraft.serverlauncher.server.Server;
import net.knarcraft.serverlauncher.server.ServerType;
import net.knarcraft.serverlauncher.userinterface.ServerConsoles;
import net.knarcraft.serverlauncher.userinterface.ServerTab;
import net.knarcraft.serverlauncher.Main;

import javax.swing.*;
import java.io.*;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Objects;
import java.util.Scanner;
import java.util.concurrent.Executors;

import static net.knarcraft.serverlauncher.Shared.downloadFile;
import static net.knarcraft.serverlauncher.Shared.readFile;
import static net.knarcraft.serverlauncher.Shared.stringBetween;

/**
 * Contains all user settings, and a list of servers.
 *
 * @author		Kristian Knarvik <kristian.knarvik@knett.no>
 * @version	0.0.0.1
 * @since		0.0.0.1
 */
public class Profile {
    private static final ArrayList<Profile> profiles = new ArrayList<>();
    private static Profile current;
    private static final String profilesDir = Main.appDir + File.separator +  "files";
    private static final String profilesFile = Main.appDir + File.separator +  "files" + File.separator + "Profiles.txt";
    private static final String jarDir = Main.appDir + File.separator + "files" + File.separator + "Jars" + File.separator;

    private final ArrayList<Collection> collections;
    private final String name;
    private boolean runInBackground;
    private int delayStartup;
    private boolean downloadJars;
    private static String vanillaVersion;
    private static String snapshotVersion;
    private static String bungeeVersion;

    private Profile(String name) {
        this.collections = new ArrayList<>();
        this.name = name;
        this.runInBackground = false;
        this.delayStartup = 0;
        this.downloadJars = false;
        profiles.add(this);
        if (current == null) {
            current = this;
        }
    }

    private Profile(String name, boolean runInBackground, int delayStartup, boolean downloadJars) {
        this.collections = new ArrayList<>();
        this.name = name;
        this.runInBackground = runInBackground;
        this.delayStartup = delayStartup;
        this.downloadJars = downloadJars;
        profiles.add(this);
        if (current == null) {
            current = this;
        }
    }

    public boolean getRunInBackground() {
        return this.runInBackground;
    }

    public int getDelayStartup() {
        return this.delayStartup;
    }

    public boolean getDownloadJars() {
        return this.downloadJars;
    }

    public static Profile getCurrent() {
        return current;
    }

    public ArrayList<Collection> getCollections() {
        return this.collections;
    }

    public String getName() {
        return this.name;
    }

    /**
     * Gets a Collection object by name.
     *
     * @param name  The name of the collection.
     * @return      A collection object.
     */
    public Collection getCollection(String name) {
        for (Collection collection : this.collections) {
            if (collection.getName().equals(name)) {
                return collection;
            }
        }
        return null;
    }

    public static Profile getProfile(String name) {
        for (Profile profile : profiles) {
            if (profile.name.equals(name)) {
                return profile;
            }
        }
        return null;
    }

    public static ArrayList<Profile> getProfiles() {
        return profiles;
    }

    public void setRunInBackground(boolean value) {
        this.runInBackground = value;
    }

    public void setDelayStartup(int value) {
        if (value >= 0) {
            this.delayStartup = value;
        }
    }

    public void setDownloadJars(boolean value) {
        this.downloadJars = value;
    }

    public static  void setCurrent(String name) {
        for (Profile profile : profiles) {
            if (profile.name.equals(name)) {
                current = profile;
                break;
            }
        }
    }

    /**
     * Adds a collection to the profile if the name is valid.
     *
     * @param name  The name of the collection and its elements.
     */
    public void addCollection(String name) {
        if (name == null) { //If a user cancels or crosses out window
            return;
        }
        if (getCollection(name) == null &&
                !name.equals("") &&
                !name.equals("All") &&
                !name.contains("!") &&
                !name.contains("?") &&
                !name.contains(";")
                ) {
            collections.add(new Collection(name));
        } else {
            JOptionPane.showMessageDialog(
                    null,
                    "A server name must my unique and not empty or \"All\"." +
                            "It can't contain any of the characters \"!\", \"?\" or \";\".",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
            );
        }
    }

    /**
     * Adds a profile if the name is valid and unique.
     *
     * @param name  The name of the new profile.
     */
    public static void addProfile(String name) {
        if (name == null) { //If a user cancels or crosses out window
            return;
        }
        if (name.equals("") && !name.contains("!") && !name.contains("?") && !name.contains(";")) {
            JOptionPane.showMessageDialog(
                    null,
                    "Profile name can't be blank.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
            );
            return;
        }
        for (Profile profile : profiles) {
            if (profile.name.equals(name)) {
                JOptionPane.showMessageDialog(
                        null,
                        "There is already a profile with this name.",
                        "Error",
                        JOptionPane.ERROR_MESSAGE
                );
                return;
            }
        }
        new Profile(name);
    }

    public void removeCollection(String name) {
        for (int i = 0; i < collections.size(); i++) {
            if (collections.get(i).getName().equals(name)) {
                this.collections.remove(i);
                Main.gui().update();
                break;
            }
        }
    }

    public static void removeProfile(String name) {
        if (profiles.size() > 1) {
            for (int i = 0; i < profiles.size(); i++) {
                if (profiles.get(i).name.equals((name))) {
                    profiles.remove(i);
                }
            }
        }
    }

    public void updateConsoles() {
        JTabbedPane consolesTab = ServerConsoles.getTab();
        consolesTab.removeAll();
        for (Collection collection : collections) {
            consolesTab.add(collection.getName(), collection.getServerConsole().getPanel());
        }
    }

    /**
     * Sends a command to a server, or all servers
     *
     * @param serverName    The target server
     * @param command       The command to send.
     */
    public void sendCommand(String serverName, String command) {
        if (serverName.equals("All")) {
            for (Collection collection : this.collections) {
                try {
                    collection.getServer().sendCommand(command);
                } catch (IOException e) {
                    JOptionPane.showMessageDialog(
                            null,
                            "Server " + collection.getName() + " caused an exception.",
                            "Error",
                            JOptionPane.ERROR_MESSAGE
                    );
                }
            }
        } else {
            Collection collection = getCollection(serverName);
            if (collection != null) {
                Server target = collection.getServer();
                try {
                    target.sendCommand(command);
                } catch (IOException e) {
                    JOptionPane.showMessageDialog(
                            null,
                            "Server " + target.getName() + " caused an exception.",
                            "Error",
                            JOptionPane.ERROR_MESSAGE
                    );
                }
            } else {
                JOptionPane.showMessageDialog(
                        null,
                        "Server " + serverName + " is invalid.",
                        "Error", JOptionPane.ERROR_MESSAGE
                );
            }
        }
    }

    /**
     * Reads all server tabs, and saves it to the variables of the corresponding servers.
     * Saves all profiles and servers to a text file.
     */
    public void save() throws FileNotFoundException {
        for (Collection collection : this.collections) {
            Server server = collection.getServer();
            ServerTab serverTab = collection.getServerTab();
            server.setPath(serverTab.getPath());
            server.setMaxRam(serverTab.getMaxRam());
            server.setType(ServerType.getByName(serverTab.getType()));
            try {
                server.setServerVersion(serverTab.getVersion());
            } catch (IllegalArgumentException e) {
                JOptionPane.showMessageDialog(
                        null,
                        "Invalid server version for " + server.getName(),
                        "Error",
                        JOptionPane.ERROR_MESSAGE
                );
            }
            server.toggle(serverTab.enabled());
        }
        if (!new File(profilesDir).exists() && !new File(profilesDir).mkdirs()) {
            JOptionPane.showMessageDialog(
                    null,
                    "Unable to create the folder " + profilesDir,
                    "Error",
                    JOptionPane.ERROR_MESSAGE
            );
            throw new FileNotFoundException("Unable to create the profiles folder: " + profilesDir);
        }
        try (PrintWriter file = new PrintWriter(profilesFile)) {
            file.println(String.format("%s;%s;%s;%s", current.name, vanillaVersion, snapshotVersion, bungeeVersion));
            file.close();
            for (Profile profile : profiles) {
                StringBuilder saveString = new StringBuilder(String.format(
                        "%s;%b;%d;%b?",
                        profile.name,
                        profile.runInBackground,
                        profile.delayStartup,
                        profile.downloadJars)
                );
                for (Collection collection : profile.collections) {
                    Server server = collection.getServer();
                    saveString.append(String.format(
                            "%s;%s;%b;%s;%s;%s;%s;%s;%s;%s!",
                            server.getName(),
                            server.getPath(),
                            server.isEnabled(),
                            server.getTypeName(),
                            server.getServerVersion(),
                            server.getMaxRam(),
                            server.getVanillaVersion(),
                            server.getSnapshotVersion(),
                            server.getSpongeVanillaVersion(),
                            server.getBungeeVersion()
                            )
                    );
                }
                saveString = new StringBuilder(saveString.substring(0, saveString.length() - 1));
                try (PrintWriter fileAppend = new PrintWriter(new FileWriter(
                        profilesFile,
                        true
                ))) {
                    fileAppend.println(saveString);
                } catch (IOException e) {
                    if (Main.gui() != null) {
                        JOptionPane.showMessageDialog(
                                null,
                                "Unable to save to file. Try running the software as an administrator.",
                                "Error",
                                JOptionPane.ERROR_MESSAGE
                        );
                        throw new FileNotFoundException("Unable to save to the profiles file.");
                    }
                }
            }
        } catch (IOException e) {
            if (Main.gui() != null) {
                JOptionPane.showMessageDialog(
                        null,
                        "Unable to save to file. Try running the software as an administrator.",
                        "Error",
                        JOptionPane.ERROR_MESSAGE
                );
            }
            throw new FileNotFoundException("Unable to create the profiles file");
        }
    }

    /**
     * Reads profiles and servers from a text file.
     */
    public static void load() {
        try (Scanner in = new Scanner(new File(profilesFile))) {
            try {
                String[] staticData = in.nextLine().split(";", -1);
                String profileName = staticData[0];
                vanillaVersion = staticData[1];
                snapshotVersion = staticData[2];
                bungeeVersion = staticData[3];
                while (in.hasNextLine()) {
                    String line = in.nextLine();
                    if (line.contains("?")) {
                        String[] data = line.split("\\?");
                        String[] profileData = data[0].split(";", -1);
                        Profile profile = parseProfile(profileData);
                        if (data[1].contains("!")) {
                            String[] servers = data[1].split("!", -1);
                            for (String server : servers) {
                                String[] serverData = server.split(";", -1);
                                parseServer(profile, serverData);
                            }
                        } else {
                            String[] serverData = data[1].split(";", -1);
                            parseServer(profile, serverData);
                        }
                    } else {
                        String[] profileData = line.split(";", -1);
                        parseProfile(profileData);
                    }
                }
                current = getProfile(profileName);
            } catch (ArrayIndexOutOfBoundsException | NumberFormatException e) {
                e.printStackTrace();
                JOptionPane.showMessageDialog(
                        null,
                        "Invalid Profile.txt file. Profiles could not be loaded. If this error persists, please manually delete the file.",
                        "Error",
                        JOptionPane.ERROR_MESSAGE
                );
                System.exit(1);
            }
            if (profiles.size() == 0) {
                addProfile("Default");
            }
        } catch (FileNotFoundException e) {
            JOptionPane.showMessageDialog(
                    null,
                    "A profiles file was not found. Default profile was created.",
                    "Info",
                    JOptionPane.INFORMATION_MESSAGE
            );
            addProfile("Default");
        }
        Main.gui().update();
        Main.gui().updateProfiles();
        current.updateConsoles();
        if (current.runInBackground) {
            Main.gui().hide();
            Executors.newSingleThreadExecutor().execute(Server::startServers);
        }
        if (current.downloadJars) {
            Executors.newSingleThreadExecutor().execute(() -> {
                try {
                    downloadJars();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            });
        }
    }

    private static Profile parseProfile(String[] profileData) {
        return new Profile(
                profileData[0],
                Boolean.parseBoolean(profileData[1]),
                Integer.parseInt(profileData[2]),
                Boolean.parseBoolean(profileData[3])
        );
    }

    private static void parseServer(Profile profile, String[] serverData) {
        profile.collections.add(new Collection(
                serverData[0],
                serverData[1],
                Boolean.parseBoolean(serverData[2]),
                serverData[3],
                serverData[4],
                serverData[5],
                serverData[6],
                serverData[7],
                serverData[8],
                serverData[9])
        );
    }

    public static void downloadJars() throws IOException {
        if (!new File(jarDir).exists() && !new File(jarDir).mkdirs()) {
            JOptionPane.showMessageDialog(
                    null,
                    "Could not create the Jars folder. Please run the program with admin permissions, or move it to a writable directory.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
            );
            throw new FileNotFoundException("Unable to create jars folder");
        }
        downloadSimple("Spigot");
        downloadSimple("Craftbukkit");
        downloadSimple("MCPCplus");
        downloadMixed("Vanilla");
        downloadMixed("Snapshot");
        downloadSponge();
        downloadBungee();
        Main.gui().setStatus("Finished downloading jars");
    }

    private static void downloadSimple(String typeName) throws FileNotFoundException {
        ServerType type = ServerType.getByName(typeName);
        String url = Objects.requireNonNull(type).getDownloadURL();
        String name = type.getName();
        Boolean success;
        for (String version : type.getVersions()) {
            File file = new File(jarDir + type.getName() + version + ".jar");
            if (!file.isFile()) {
                Path filePath = Paths.get(jarDir + type.getName() + version + ".jar");
                if (Main.gui() != null) {
                    Main.gui().setStatus("Downloading: " + name + version + ".jar");
                }
                success = downloadFile(url + name + version + ".jar", filePath);
                if (!success) {
                    if (Main.gui() != null) {
                        Main.gui().setStatus("Error downloading: " + name + version + ".jar");
                    }
                    throw new FileNotFoundException("Error downloading: " + name + version + ".jar");
                }
            }
        }
    }

    private static void downloadMixed(String typeName) throws IOException {
        AdvancedServerType type = (AdvancedServerType) ServerType.getByName(typeName);
        String url = Objects.requireNonNull(type).getDownloadURL();
        String name = type.getName();
        String versionText;
        String newestVersion;
        Boolean success;
        for (String version : type.getVersions()) {
            File file = new File(jarDir + type.getName() + version + ".jar");
            Path filePath = Paths.get(jarDir + type.getName() + version + ".jar");
            if (Main.gui() != null) {
                Main.gui().setStatus("Downloading: " + name + version + ".jar");
            }
            if (version.equals("Latest")) {
                try {
                    versionText = readFile(type.getVersionURL());
                } catch (IOException e) {
                    throw new IOException("Error reading: " + type.getVersionURL());
                }
                newestVersion = stringBetween(versionText, type.getSrcStart(), type.getSrcEnd());
                if (!file.isFile() || !newestVersion.equals(getVersion(name))) {
                    success = downloadFile(
                            url + newestVersion + type.getDownloadURLPart() + newestVersion + ".jar",
                            filePath
                    );
                    setVersion(name, newestVersion);
                    if (!success) {
                        if (Main.gui() != null) {
                            Main.gui().setStatus("Error downloading: " + name + version + ".jar");
                        }
                        throw new FileNotFoundException("Error downloading: " + name + version + ".jar");
                    }
                }
            } else {
                if (!file.isFile()) {
                    success = downloadFile(url + version + type.getDownloadURLPart() + version + ".jar", filePath);
                    if (!success) {
                        if (Main.gui() != null) {
                            Main.gui().setStatus("Error downloading: " + name + version + ".jar");
                        }
                        throw new FileNotFoundException("Error downloading: " + name + version + ".jar");
                    }
                }
            }
        }
    }

    private static void downloadSponge() throws IOException {
        AdvancedServerType type = (AdvancedServerType) ServerType.getByName("SpongeVanilla");
        String url = Objects.requireNonNull(type).getDownloadURL();
        String name = type.getName();
        String versionText;
        String newestVersion;
        Boolean success;
        for (String version : type.getVersions()) {
            File file = new File(jarDir + name + version + ".jar");
            Path filePath = Paths.get(jarDir + name + version + ".jar");
            if (Main.gui() != null) {
                Main.gui().setStatus("Downloading: " + name + version + ".jar");
            }
            try {
                versionText = readFile(type.getVersionURL() + version);
            } catch (IOException e) {
                throw new IOException("Error reading: " + type.getVersionURL());
            }
            newestVersion = stringBetween(versionText, type.getSrcStart(), type.getSrcEnd());
            if (!file.isFile()) {
                success = downloadFile(
                        url + newestVersion + type.getDownloadURLPart() + newestVersion + ".jar",
                        filePath
                );
                if (!success) {
                    if (Main.gui() != null) {
                        Main.gui().setStatus("Error downloading: " + name + version + ".jar");
                    }
                    throw new FileNotFoundException("Error downloading: " + name + version + ".jar");
                }
            }
        }
    }

    private static void downloadBungee() throws IOException {
        AdvancedServerType type = (AdvancedServerType) ServerType.getByName("Bungee");
        String url = Objects.requireNonNull(type).getDownloadURL();
        String name = type.getName();
        String versionText;
        String newestVersion;
        Boolean success;
        File file = new File(jarDir + type.getName() + ".jar");
        Path filePath = Paths.get(jarDir + type.getName() + ".jar");
        if (Main.gui() != null) {
            Main.gui().setStatus("Downloading: " + name + ".jar");
        }
        try {
            versionText = readFile(type.getVersionURL());
        } catch (IOException e) {
            throw new IOException("Error reading: " + type.getVersionURL());
        }
        newestVersion = stringBetween(versionText, type.getSrcStart(), type.getSrcEnd());
        if (!file.isFile() || !newestVersion.equals(getVersion(name))) {
            success = downloadFile(url, filePath);
            setVersion(name, newestVersion);
            if (!success) {
                if (Main.gui() != null) {
                    Main.gui().setStatus("Error downloading: " + name + ".jar");
                }
                throw new FileNotFoundException("Error downloading: " + name + ".jar");
            }
        }
    }

    /**
     * Returns the current version of a type
     *
     * @param type  The version type
     * @return      The version string
     */
    private static String getVersion(String type) {
        switch (type) {
            case "Vanilla":
                return vanillaVersion;
            case "Snapshot":
                return snapshotVersion;
            case "Bungee":
                return bungeeVersion;
            default:
                return "";
        }
    }

    /**
     * Sets a server type's last downloaded version.
     *
     * @param type      The version type
     * @param version   The version string
     */
    private static void setVersion(String type, String version) {
        if (!type.equals("")) {
            switch (type) {
                case "Vanilla":
                    vanillaVersion = version;
                    break;
                case "Snapshot":
                    snapshotVersion = version;
                    break;
                case "Bungee":
                    bungeeVersion = version;
            }
        }
    }
}
